from musurgia.agpdf.drawobject import DrawObject
from musurgia.agpdf.labeled import Labeled
from musurgia.agpdf.markline import MarkLine
from musurgia.agpdf.named import Named


class Line(DrawObject, Labeled, Named):
    def __init__(self, length, factor=1, line_distance=20, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._length = None

        self._line_distance = None
        self.line_distance = line_distance
        self._factor = None
        self.length = length
        self.factor = factor
        self._start_mark_line = MarkLine()
        self._end_mark_line = MarkLine(show=False)
        self._x1 = None
        self._x2 = None
        self._y1 = None
        self._y2 = None
        self._page = None

    @property
    def x1(self):
        return self._x1

    @property
    def x2(self):
        return self._x2

    @property
    def y1(self):
        return self._y1

    @property
    def y2(self):
        return self._y2

    @property
    def page(self):
        return self._page

    @property
    def factor(self):
        if self._factor is None:
            self._factor = 1
        return self._factor

    @factor.setter
    def factor(self, val):
        self._factor = val

    @property
    def line_distance(self):
        return self._line_distance

    @line_distance.setter
    def line_distance(self, val):
        self._line_distance = val

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, val):
        self._length = val

    @property
    def actual_length(self):
        return self.length * self.factor

    @property
    def start_mark_line(self):
        self._start_mark_line.relative_x = self.relative_x
        self._start_mark_line.relative_y = self.relative_y
        return self._start_mark_line

    @property
    def end_mark_line(self):
        self._end_mark_line.relative_x = self.relative_x + self.length * self.factor
        self._end_mark_line.relative_y = self.relative_y
        return self._end_mark_line

    def get_relative_x2(self):
        return self.relative_x + self.actual_length

    def get_relative_y2(self):
        return max([self.start_mark_line.get_relative_y2(), self.end_mark_line.get_relative_y2(), self.line_distance])

    def draw(self, pdf):
        x1 = pdf.x + self.relative_x
        x2 = pdf.x + self.relative_x + self.actual_length
        y1 = pdf.y + self.relative_y
        y2 = y1

        self._x1, self._x2, self._y1, self._y2 = x1, x2, y1, y2
        self._page = pdf.page
        if self.name:
            self.name.draw(pdf)

        if self.show:
            self.start_mark_line.draw(pdf)
            self.end_mark_line.draw(pdf)
            pdf.line(x1=x1, y1=y1, x2=x2, y2=y2)
            for text_label in self._text_labels:
                text_label.draw(pdf)

        pdf.x = x2



class FontError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Font(object):
    _FAMILY = ['Arial']
    _WEIGHT = ['bold', 'regular']
    _STYLE = ['italic', 'regular']

    def __init__(self, family='Arial', weight='regular', style='regular', size=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._family = None
        self._weight = None
        self._style = None
        self._size = None

        self.family = family
        self.weight = weight
        self.style = style
        self.size = size

    @property
    def family(self):
        return self._family

    @family.setter
    def family(self, val):
        if val not in self._FAMILY:
            raise FontError('{} not a valid value: {}'.format(val, self._FAMILY))
        self._family = val

    @property
    def weight(self):
        return self._weight

    @weight.setter
    def weight(self, val):
        if val not in self._WEIGHT:
            raise FontError('{} not a valid value: {}'.format(val, self._WEIGHT))
        self._weight = val

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, val):
        if val not in self._STYLE:
            raise FontError('{} not a valid value: {}'.format(val, self._STYLE))
        self._style = val

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, val):
        self._size = val



from unittest import TestCase

from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.quantize import get_quantized_values


class Test(TestCase):
    def setUp(self) -> None:
        self.arith_prog = ArithmeticProgression()

    def test_1(self):
        result = {'a': None, 'an': None, 'n': None, 'd': None, 's': None}
        self.assertEqual(self.arith_prog._parameters_dict, result)

    def test_2(self):
        self.arith_prog.n = 15
        self.arith_prog.a1 = 1
        self.arith_prog.d = 2
        result = {'a': 1, 'an': 29, 'n': 15, 'd': 2, 's': 225.0}
        self.assertEqual(self.arith_prog.parameters_dict, result)

    def test_3(self):
        self.arith_prog.n = 15
        self.arith_prog.a1 = 1
        self.arith_prog.d = 2
        result = [1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29]
        self.assertEqual(list(self.arith_prog), result)

    def test_4(self):
        self.arith_prog.a1 = 36
        self.arith_prog.an = 25
        self.arith_prog.s = 230
        # print(self.arith_prog.parameters_dict)
        values = list(self.arith_prog)
        # print(values)
        quantized = get_quantized_values(values, 1)
        # print(quantized)
        # print(sum(quantized))
        result = [36, 34, 32, 31, 29, 26, 25]
        self.assertEqual(quantized, result)



import os
from unittest import TestCase

from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.fractaltree.fractalmusicsquare import Square
from musurgia.testcomparefiles import TestCompareFiles

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.square = Square(duration=600, tree_permutation_order=[4, 1, 2, 3], proportions=[1, 2, 3, 4])
        # for module in self.square.modules.values():
        #     module.tempo = 72
        tempi = [round(tempo) for tempo in ArithmeticProgression(a1=40, an=72, n=4)]
        for row in self.square.rows:
            for module in row.modules:
                tempo = tempi[row.number - 1]
                module.tempo = tempo

    def test_1(self):
        text_path = path + '_test_1.txt'
        self.square.write_infos(text_path=text_path, show_module_tempo=True, show_quarter_durations=True)
        TestCompareFiles().assertTemplate(text_path)

    def test_2(self):
        text_path = path + '_test_2.txt'
        module = self.square.get_module(1, 1)
        module.duration = 3

        self.square.write_infos(text_path=text_path, show_module_tempo=True,
                                show_quarter_durations=True)
        TestCompareFiles().assertTemplate(text_path)

    def test_3(self):
        self.square.change_module_quarter_duration(1, 1, 3)

        text_path = path + '_test_3.txt'

        self.square.write_infos(text_path=text_path, show_module_tempo=True,
                                show_quarter_durations=True)

        TestCompareFiles().assertTemplate(text_path)

    def test_4(self):
        self.square = Square(duration=600, tree_permutation_order=[3, 4, 2, 1], proportions=[1, 2, 3, 4],
                             reading_direction='vertical')
        self.square.change_module_duration(1, 1, 3)

        orders = [module.permutation_order for module in self.square.get_all_modules()]
        result = [[4, 1, 2, 3], [3, 2, 1, 4], [1, 3, 4, 2], [2, 4, 3, 1], [2, 3, 1, 4], [1, 4, 2, 3], [4, 2, 3, 1],
                  [3, 1, 4, 2], [1, 4, 3, 2], [2, 3, 4, 1], [3, 1, 2, 4], [4, 2, 1, 3], [3, 2, 4, 1], [4, 1, 3, 2],
                  [2, 4, 1, 3], [1, 3, 2, 4]]

        self.assertEqual(orders, result)



import itertools
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from prettytable import PrettyTable
from quicktions import Fraction

from musurgia import basic_functions, scaledvalues
from musurgia.fractaltree.fractaltree import FractalTree
from musurgia.fractaltree.midigenerators import RelativeMidi, MidiGenerator
from musurgia.permutation import permute
from musurgia.quantize import get_quantized_values
from musurgia.timing import Timing


class FractalMusicException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class SetDurationFirstException(FractalMusicException):
    def __init__(self, *args):
        super().__init__('set duration first!', *args)


class SetTempoFirstException(FractalMusicException):
    def __init__(self, *args):
        super().__init__('set tempo first!', *args)


class TempoIsAlreadySet(FractalMusicException):
    def __init__(self, *args):
        super().__init__('FractalMusic().tempo can only be set once', *args)


class ChildTempoIsAlreadySet(FractalMusicException):
    def __init__(self, *args):
        super().__init__('FractalMusic().tempo of parent can not be set after setting tempo of child', *args)


class MergeException(FractalMusicException):
    def __init__(self, *args):
        super().__init__(*args)


class MergeTempoException(MergeException):
    def __init__(self, tempo, *args):
        super().__init__('nodes to merge must have the same tempo {}'.format(tempo), *args)


class FractalMusic(FractalTree):
    def __init__(self,
                 midi_generator=None,
                 duration=None,
                 tempo=None,
                 quarter_duration=None,
                 tree_directions=None,
                 permute_directions=True,
                 *args, **kwargs):

        # super().__init__(value=duration, *args, **kwargs)
        super().__init__(*args, **kwargs)

        self._midi_value = None
        self._chord = None
        self._midi_generator = None
        self._children_generated_midis = None
        self._tree_directions = None
        self._permute_directions = True
        self._tempo = None

        self._corrected_score_duration = None

        self.midi_generator = midi_generator
        self.duration = duration
        self.tempo = tempo
        self.permute_directions = permute_directions
        self.tree_directions = tree_directions
        self.quarter_duration = quarter_duration

    @property
    def duration(self):
        return self.value

    @duration.setter
    def duration(self, val):
        if val and not isinstance(val, Fraction):
            val = Fraction(val)
        self.value = val

    @property
    def quarter_duration(self):
        if not self.tempo:
            raise SetTempoFirstException()
        try:
            return Fraction(self.duration * self.tempo, 60)
        except TypeError:
            return Fraction(Fraction(self.duration * self.tempo), Fraction(60))

    @quarter_duration.setter
    def quarter_duration(self, val):
        if val is not None:
            if not self.tempo:
                raise SetTempoFirstException()
            try:
                self.duration = Fraction(val * 60, self.tempo)
            except TypeError:
                self.duration = Fraction(Fraction(val * 60), Fraction(self.tempo))

    def _child_has_tempo(self):
        children_tempi = [child.tempo for child in self.get_children() if child.tempo]
        if children_tempi:
            return True
        else:
            return False

    @property
    def tempo(self):
        return self._tempo

    @tempo.setter
    def tempo(self, val):
        if self._tempo:
            raise TempoIsAlreadySet()
        if self._child_has_tempo():
            raise ChildTempoIsAlreadySet()
        self._tempo = val
        for child in self.get_children():
            child.tempo = val

    def set_non_tempi(self, val=60):
        try:
            self.tempo = val
        except TempoIsAlreadySet:
            pass
        except ChildTempoIsAlreadySet:
            for child in self.get_children():
                child.set_non_tempi(val=val)

    @property
    def quarter_position_in_tree(self):
        return self.position_in_tree * self.tempo / 60

    @property
    def permute_directions(self):
        return self._permute_directions

    @permute_directions.setter
    def permute_directions(self, value):
        if not isinstance(value, bool):
            raise TypeError()
        self._permute_directions = value

    @property
    def tree_directions(self):
        return self._tree_directions

    # todo: setting tree_direction after setting midi_range has no effect
    @tree_directions.setter
    def tree_directions(self, val):
        if val is None:
            val = [1, -1]
        length = len(self.permutation_order)
        if len(val) > length:
            val = val[:length]
        elif len(val) < length:
            output = []
            cycled = itertools.cycle(val)
            for i in range(length):
                output.append(cycled.__next__())
            val = output
        self._tree_directions = val

    @property
    def midi_generator(self):
        if self._midi_generator is None:
            try:
                if not self.duration:
                    raise SetDurationFirstException()
                self._midi_generator = RelativeMidi(midi_range=[71, 71], proportions=self.children_fractal_values,
                                                    directions=None)
                self._midi_generator.node = self
            except AttributeError:
                raise Exception('midi_generator is None and cannot be set:')
        else:
            self._midi_generator.node = self

        if isinstance(self._midi_generator, RelativeMidi):

            if not self._midi_generator._directions:
                tree_directions = self.get_root().tree_directions

                if self.permute_directions:
                    directions = permute(tree_directions, self.permutation_order)
                else:
                    directions = tree_directions

                self._midi_generator.directions = directions

            if self._midi_generator.midi_range is None:
                if self.is_root:
                    self._midi_generator.midi_range = [71]
                else:
                    self._midi_generator.midi_range = self.auto_midi_range
                    self._midi_generator._auto_ranged = True

        return self._midi_generator

    @midi_generator.setter
    def midi_generator(self, value):
        if value is not None:
            if not (isinstance(value, MidiGenerator)):
                err = 'midi_generator can only be an instance of subclasses of MidiGenerator or None. None=RelativeMidi(midi_range=self.midi_range, proportions=permute(self.fractal_proportions, self.fractal.permutation_order)'
                raise TypeError(err)

        self._midi_generator = value
        if self._midi_generator:
            self._midi_generator.node = self

    @property
    def _midi_iterator(self):
        mg = self.midi_generator
        return mg.iterator

    @_midi_iterator.setter
    def _midi_iterator(self, value):
        raise AttributeError('_midi_iterator cannot be set directly. Use midi-generator!')

    @property
    def children_generated_midis(self):
        if self._children_generated_midis is None:
            self._children_generated_midis = []
            self._children_generated_midis = list(self._midi_iterator)

        return self._children_generated_midis

    @property
    def _children_midis(self):
        return [child.midi_value for child in self.get_children()]

    @property
    def midi_value(self):
        if self._midi_value is None:
            if self.is_root:
                self._midi_value = 71
            else:
                self._midi_value = self.up.children_generated_midis[self.up.get_children().index(self)]
        return self._midi_value

    @midi_value.setter
    def midi_value(self, value):
        if value is not None:
            if isinstance(value, int) or isinstance(value, float):
                if value >= 18 or value == 0:
                    self._midi_value = value
                else:
                    raise ValueError('midi can be 0 and greater than 18')
            else:
                raise TypeError('midi can only be int, float or None not {}'.format(type(value)))
        else:
            self._midi_value = value

    @property
    def auto_midi_range(self):
        if self.is_root:
            raise AttributeError('root has no auto_midi_range')

        parent_midis = self.up.children_generated_midis
        self_index = self.up.get_children().index(self)
        auto_range = parent_midis[self_index:self_index + 2]
        return auto_range

    def get_choral_midis(self, range_factor=1, direction=None, last=False):
        range_ = (self.midi_generator.midi_range[1] - self.midi_generator.midi_range[0]) * range_factor
        self_direction = range_ / abs(range_)
        if direction is None or direction == 0:
            pass
        elif direction == 1:
            range_ = abs(range_)
        elif direction == -1:
            range_ = -abs(range_)

        children_midis = self.children_generated_midis
        choral_midis = []

        step = 2 / self.midi_generator.microtone
        if self_direction > 0:
            scale = scaledvalues.ScaledValues(min(children_midis), max(children_midis), self.midi_value,
                                              self.midi_value + range_,
                                              step=step)
        else:
            scale = scaledvalues.ScaledValues(max(children_midis), min(children_midis), self.midi_value,
                                              self.midi_value + range_,
                                              step=step)
        if not last:
            min_midi = min(children_midis)
            max_midi = max(children_midis)
            if self.midi_value == min_midi:
                children_midis.remove(max_midi)
            else:
                children_midis.remove(min_midi)
        for midi in children_midis:
            new_midi = scale(midi)
            choral_midis.append(new_midi)

        choral_midis = sorted(choral_midis)
        choral_midis = list(set(choral_midis))
        return choral_midis

    def get_choral(self, range_factor=1, direction=None, last=False):
        choral_midis = self.get_choral_midis(range_factor=range_factor, direction=direction, last=last)
        return TreeChord(quarter_duration=self.quarter_duration, midis=choral_midis)

    @property
    def chord(self):
        if self._chord is None:  # or self._note.midis!=[self.midi] or self._note.duration!=self.duration:
            self._chord = TreeChord(quarter_duration=self.quarter_duration, midis=[self.midi_value])
        else:
            self._chord.quarter_duration = self.quarter_duration

        return self._chord

    def set_chord(self, chord):
        self._chord = chord

    def get_simple_format(self, layer=None):
        if layer is None:
            layer = self.get_farthest_leaf().get_distance()
        self.get_layer(layer=layer)

        simple_format = SimpleFormat()
        for chord in [node.chord for node in basic_functions.flatten(self.get_layer(layer))]:
            copied_chord = chord.__deepcopy__()
            simple_format.add_chord(copied_chord)

        return simple_format

    def split(self, *proportions):
        if hasattr(proportions[0], '__iter__'):
            proportions = proportions[0]

        proportions = [Fraction(prop) for prop in proportions]

        for prop in proportions:
            duration = self.quarter_duration * prop / sum(proportions)
            new_node = self.copy()
            new_node.multi = self.multi
            new_node.quarter_duration = duration
            new_node._fractal_order = self.fractal_order
            new_node.midi_value = self.midi_value
            new_node.midi_generator.midi_range = self.midi_generator.midi_range
            if self._chord:
                new_node.set_chord(self.chord.__deepcopy__())
            self.add_child(new_node)

        return self.get_children()

    def _check_merge_nodes(self, nodes):
        tempo = set([node.tempo for node in nodes])
        if len(tempo) != 1:
            raise MergeTempoException(tempo)

    # def merge_children(self, *lengths):
    #     super().merge_children(*lengths)

    def _refill_duration(self):
        if not self._value:
            self._value = sum([child._refill_duration() for child in self.get_children()])
        else:
            return self._value

    def quantize_leaves(self, grid_size):
        def _reset_durations():
            for node in self.traverse():
                node._value = None

        #     quantizing quarter_durations!
        leaves = list(self.traverse_leaves())
        durations = [leaf.quarter_duration for leaf in leaves]
        quantized_durations = get_quantized_values(durations, grid_size)
        if quantized_durations:
            _reset_durations()
        for leaf, quantized_duration in zip(leaves, quantized_durations):
            leaf.quarter_duration = quantized_duration
        self._refill_duration()

    def change_midis(self):
        if not isinstance(self._midi_generator, RelativeMidi):
            raise TypeError(
                'set_reduced_auto_ranges can only be applied to FractalMusic nodes with RelativeMidi as midi_generator')
        else:

            # print(self.children_generated_midis)
            directions = self.midi_generator.directions
            midi_range = self.midi_generator.midi_range
            microtone = self.midi_generator.microtone

            # children = iter(self.get_children())
            # old_generated_midis = self.children_generated_midis
            # print("old_generated_midis", old_generated_midis)
            # self._children_generated_midis = [node.midi_value for node in self.get_children()]
            # self._children_generated_midis.append(old_generated_midis[-1])
            # print("new_generated_midis", old_generated_midis)

            self.midi_generator = None
            self.midi_generator.proportions = None
            self.midi_generator.midi_range = midi_range
            self.midi_generator.microtone = microtone
            self.midi_generator.directions = directions

            self._children_generated_midis = None

            for child in self.get_children():
                child.midi_value = None
                child._chord = None

            for child in self.get_children():
                if isinstance(child.midi_generator, RelativeMidi):
                    child.midi_generator.midi_range = None
                    child.midi_generator.directions = None
                    child.midi_generator.proportions = None
                    child.midi_generator._iterator = None
                    child._children_generated_midis = None

    def get_score_template(self):
        score = TreeScoreTimewise()
        score.tuplet_line_width = 2.4
        score.page_style.orientation = 'landscape'
        score.page_style.system_distance = 180
        score.page_style.staff_distance = 150
        score.page_style.top_system_distance = 150
        score.page_style.bottom_margin = 100

        score.add_title('module: {}'.format(self.__name__))
        clock = Timing.get_clock(round(self.duration), mode='msreduced')
        score.add_subtitle(
            'duration: {}'.format(clock))
        score.accidental_mode = 'modern'
        return score

    def get_score(self, score=None, layer_number=None, show_fractal_orders=False, show_midis=None,
                  barline='light-heavy',
                  show_metronome=True):
        if not score:
            score = self.get_score_template()
        score.set_time_signatures(durations=self.quarter_duration)
        if show_fractal_orders:
            for node in self.traverse():
                node.chord.add_lyric(node.fractal_order)

        if show_midis:
            for node in self.traverse():
                midi_value = node.midi_value
                if int(midi_value) == midi_value:
                    midi_value = int(midi_value)
                node.chord.add_words(midi_value, enclosure='none', relative_y=10)

        def layer_to_score(layer_number, part_number):
            try:
                sf = self.get_simple_format(layer_number)
                sf.auto_clef()
                v = sf.to_stream_voice(1)
                v.add_to_score(score, 1, part_number)
            except ValueError:
                print('module {}: number_of_layers={}: getting layer {} not possible'.format(self.__name__,
                                                                                             self.number_of_layers,
                                                                                             i))

        if layer_number is None:
            for i, j in zip(range(self.number_of_layers + 1), range(1, self.number_of_layers + 2)):
                layer_to_score(i, j)

        elif hasattr(layer_number, '__iter__'):
            for i, j in zip(layer_number, range(1, len(layer_number) + 1)):
                layer_to_score(i, j)
        else:
            layer_to_score(layer_number, 1)

        if show_metronome:
            score.get_measure(1).get_part(1).add_metronome(per_minute=self.tempo, relative_y=40)
        score.get_measure(-1).set_barline_style(barline)
        return score

    def get_root_score(self, score=None, layer_number=None, show_fractal_orders=False, show_midis=False
                       # , show_positions=False
                       ):

        if not score:
            score = self.get_score_template()

        try:
            return self.get_score(score=score, layer_number=layer_number, show_fractal_orders=show_fractal_orders,
                                  show_midis=show_midis)
        except SetTempoFirstException as err:
            if not self.get_children():
                raise err
            else:
                for child in self.get_children():
                    score.extend(
                        child.get_score(layer_number=layer_number, show_fractal_orders=show_fractal_orders,
                                        show_midis=show_midis, barline='light-light'))

        # if set_time_signatures:
        #     if layer_number == 0 or not self.get_children():
        #         score.set_time_signatures(durations=self.quarter_duration, barline_style='light-light', times=times)
        #     else:
        #         durations = [child.quarter_duration for child in self.get_children() if child.quarter_duration != 0]
        #         score.set_time_signatures(
        #             durations=durations, barline_style='light-light', times=times)
        # else:

        score.accidental_mode = 'modern'

        # if show_positions:
        #     for node in self.get_children()[1:]:
        #         position_in_tree = float(node.position_in_tree)
        #         quarter_position_in_tree = position_in_tree * self.tempo / 60
        #         position_in_score = quarter_position_in_tree * 60 / self.score_tempo
        #         node.chord.add_words(
        #             Timing.get_clock(time=round(position_in_score, 1), mode='msreduced'),
        #             enclosure='rectangle', relative_y=40)
        score.get_measure(-1).set_barline_style('light-heavy')
        return score

    # get_score = get_fractal_score

    def write_infos(self, file_name):
        os.system('touch ' + file_name)
        file = open(file_name, 'w')
        x = PrettyTable()

        x.field_names = ["name", "f_o", "quarters", "duration", "midi", "perm_ord", "multi", "directions",
                         "midi_range", "childr_midis"]

        for node in self.traverse():
            x.add_row([node.__name__, node.fractal_order, round(float(node.quarter_duration), 2),
                       Timing(quarter_duration=float(node.quarter_duration), tempo=self.tempo).clock,
                       node.midi_value,
                       node.permutation_order, node.multi, node.midi_generator.directions,
                       node.midi_generator.midi_range, node.children_generated_midis]
                      )

        if self.is_root:
            file.write('name: root')
        else:
            file.write('name: {} fo: {}'.format(self.__name__, self.fractal_order))
        file.write('\n')
        file.write(
            'quarter_duration: {}, tempo: {}, duration: {}'.format(round(float(self.quarter_duration), 2),
                                                                   self.tempo,
                                                                   round(float(self.duration), 2)))
        file.write('\n')
        file.write('tree directions: {}'.format(self.tree_directions))
        file.write('\n')
        file.write('\n')
        file.write(x.get_string())
        file.close()

    def copy(self):
        copied_midi_generator = self.midi_generator.copy()

        if isinstance(copied_midi_generator, RelativeMidi):
            copied_midi_generator.midi_range = None
            copied_midi_generator._directions = None
            copied_midi_generator.proportions = None
            copied_midi_generator._iterator = None

        copied = super().copy()
        copied.tempo = self.tempo
        copied.midi_generator = copied_midi_generator
        copied.tree_directions = self.tree_directions
        copied.permute_directions = self.permute_directions

        if self.fertile is False:
            copied._midi_value = self._midi_value
        return copied

    def reduce_children(self, condition):
        super().reduce_children(condition)

        try:
            self.change_midis()
        except TypeError:
            pass

    def __deepcopy__(self, memodict={}):
        copied = super().copy()
        copied.multi = self.multi
        copied._fractal_order = self.fractal_order
        copied._name = self.__name__
        copied.tempo = self.tempo
        if self._midi_generator is not None:
            copied._midi_generator = self._midi_generator.__deepcopy__()
        copied.permute_directions = self.permute_directions
        copied.tree_directions = self.tree_directions

        copied._up = self.up
        for child in self.get_children():
            copied.add_child(child.__deepcopy__())

        if self.fertile is False:
            copied._midi_value = self._midi_value
        return copied



from unittest import TestCase

from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def setUp(self) -> None:
        self.ft = FractalTree(tree_permutation_order=[3, 1, 2], proportions=[1, 2, 3], value=10)

    def test_1(self):
        self.ft.add_layer()
        result = [(2, 1), (2, 2), (2, 3)]
        self.assertEqual([child.multi for child in self.ft.get_children()], result)



import os
from unittest import TestCase

from musurgia.testcomparefiles import TestCompareFiles

tfc = TestCompareFiles()
path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def test_1(self):
        file_path = path + '_a1.pdf'
        template_path = path + '_a2.pdf'
        tfc.assertTemplate(file_path, template_path)

    def test_2(self):
        with self.assertRaises(AssertionError):
            file_path = path + '_a1.pdf'
            template_path = path + '_b.pdf'
            tfc.assertTemplate(file_path, template_path)



import os

from prettytable import PrettyTable
from quicktions import Fraction

from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.fractaltree.fractaltree import FractalTreeException


class Module(FractalMusic):
    def __init__(self, row_number=None, column_number=None, *args, **kargs):
        super().__init__(*args, **kargs)
        self.row_number = row_number
        self.column_number = column_number
        self._parent_row = None
        self._parent_square = None
        self._parent_column = None
        self._name = None

    @property
    def parent_square(self):
        return self._parent_square

    @property
    def parent_row(self):
        return self._parent_row

    @property
    def parent_column(self):
        return self._parent_column

    def change_duration(self, new_duration):
        if self.parent_square:
            self.parent_square.change_module_duration(self.row_number, self.column_number, new_duration)
        else:
            raise ValueError('parent_square is not set')

    def change_quarter_duration(self, new_quarter_duration):
        if self.parent_square:
            self.parent_square.change_module_quarter_duration(self.row_number, self.column_number, new_quarter_duration)
        else:
            raise ValueError('parent_square is not set')

    def set_name(self, value):
        self._name = value

    @property
    def __name__(self):
        if self._name:
            return self._name

        if self.row_number and self.column_number:
            return str(self.row_number) + '_' + str(self.column_number)
        else:
            return super().__name__

    def write_square_infos(self, text_path, show_quarter_durations=False):
        os.system('touch ' + text_path)
        file = open(text_path, 'w')
        file.write("module_tempo: " + str(self.tempo))
        file.write("\n")
        x = PrettyTable(hrules=1)

        leaf_names = [leaf.__name__ for leaf in self.traverse_leaves()]
        leaf_fractal_orders = [leaf.fractal_order for leaf in self.traverse_leaves()]

        leaf_durations = [leaf.duration for leaf in self.traverse_leaves()]

        rounded_leaf_durations = [round(float(dur), 2) for dur in leaf_durations]

        leaf_quarter_durations = [leaf.quarter_duration for leaf in self.traverse_leaves()]
        rounded_quarter_durations = [round(float(dur), 2) for dur in leaf_quarter_durations]

        x.field_names = ["name", "info", *leaf_names, "sum"]
        x.add_row([self.__name__, 'frac_ord', *leaf_fractal_orders, " "])
        x.add_row([self.__name__, 'durations', *rounded_leaf_durations, round(float(sum(leaf_durations)), 2)])
        x.add_row(
            [self.__name__, 'quarter_dur', *rounded_quarter_durations, round(float(sum(leaf_quarter_durations)), 2)])

        file.write(x.get_string())
        file.close()

    # def get_time_line(self, factor=1, **kwargs):
    #     return TreeTimeLine(length=self.quarter_duration, factor=factor, **kwargs)

    # def get_pdf(self, factor=None):
    #     pdf = Pdf(orientation='landscape')
    #
    #     # pdf.set_margins(20, 20, 20)
    #
    #     pdf.write_line(txt=self.__name__)
    #     pdf.write_line(txt='tree_permutation_order: ' + str(self.tree_permutation_order))
    #     pdf.write_line(txt='permutation_order: ' + str(self.permutation_order))
    #
    #     pdf.write_line(txt='module_tempo: ' + str(self.module_tempo))
    #     pdf.write_line(txt='score_tempo: ' + str(self.score_tempo))
    #     pdf.write_line(txt='quarter_duration: ' + str(self.quarter_duration))
    #     pdf.write_line(txt='duration: ' + str(round(float(self.actual_duration), 2)))
    #
    #     if factor is None:
    #         factor = (pdf.w - pdf.l_margin - pdf.r_margin) / self.quarter_duration
    #
    #     module_tl = TreeTimeLine(length=self.quarter_duration, factor=factor)
    #     label_6 = Label(text=round(float(self.quarter_duration), 2), relative_y=-1,
    #                     relative_x=self.quarter_duration * module_tl.factor / 2)
    #     label_6.font.size = 5
    #     module_tl.mark_line.add_label(label_6)
    #
    #     label_7 = Label(text=round(float(self.actual_duration), 2), relative_y=1,
    #                     relative_x=self.quarter_duration * module_tl.factor / 2)
    #     label_7.font.size = 5
    #     module_tl.mark_line.add_label(label_7)
    #
    #     for child in self.get_children():
    #         child_tl = module_tl.add_child(TreeTimeLine(length=child.quarter_duration, x_offset=None, y_offset=None))
    #
    #         label_1 = Label(text=child.fractal_order)
    #         label_1.font.size = 5
    #         child_tl.mark_line.add_label(label_1)
    #
    #         label_2 = Label(text=round(float(child.quarter_position_in_tree), 2), relative_y=5)
    #         label_2.font.size = 5
    #         child_tl.mark_line.add_label(label_2)
    #
    #         label_3 = Label(text=round(float(child.position_in_score), 2), relative_y=7)
    #         label_3.font.size = 5
    #         child_tl.mark_line.add_label(label_3)
    #
    #         label_4 = Label(text=round(float(child.quarter_duration), 2), relative_y=-1,
    #                         relative_x=child.quarter_duration * child_tl.factor / 2)
    #         label_4.font.size = 5
    #         child_tl.mark_line.add_label(label_4)
    #
    #         label_5 = Label(text=round(float(child.actual_duration), 2), relative_y=1,
    #                         relative_x=child.quarter_duration * child_tl.factor / 2)
    #         label_5.font.size = 5
    #         child_tl.mark_line.add_label(label_5)
    #
    #     pdf.add_time_line(module_tl)
    #
    #     return pdf

    def __deepcopy__(self, memodict={}):
        copied = super().__deepcopy__(memodict)
        copied.row_number = self.row_number
        copied.column_number = self.column_number

        return copied


class RowColumn(object):
    def __init__(self, square, number, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._square = None
        self.square = square
        self._number = None
        self.number = number
        self._modules = []

    @property
    def square(self):
        return self._square

    @square.setter
    def square(self, value):
        if not isinstance(value, Square):
            raise TypeError('square.value must be of type Square not {}'.format(type(value)))
        self._square = value

    @property
    def number(self):
        return self._number

    @number.setter
    def number(self, value):
        if not isinstance(value, int):
            raise TypeError('number.value must be of type int not {}'.format(type(value)))
        if not 0 < value <= self.square.side_size:
            raise ValueError('number.value must be of between 0 and {}'.format(self.square.side_size))
        self._number = value

    @property
    def modules(self):
        return self._modules

    def get_module(self, number):
        return self.modules[number - 1]

    @property
    def duration(self):
        return sum([module.quarter_duration for module in self.modules])


class Row(RowColumn):
    def __init__(self, square, number, *args, **kwargs):
        super().__init__(square, number, *args, **kwargs)
        self._name = None

    def _add_module(self, module):
        module._parent_row = self
        self._modules.append(module)

    # def set_module_tempo(self, tempo):
    #     for module in self.modules:
    #         module.module_tempo = tempo
    #
    # def set_score_tempo(self, tempo):
    #     for module in self.modules:
    #         module.score_tempo = tempo

    def set_tempo(self, tempo):
        for module in self.modules:
            module.tempo = tempo

    def set_name(self, val):
        self._name = val

    @property
    def __name__(self):
        return self._name


class Column(RowColumn):
    def __init__(self, square, number, *args, **kwargs):
        super().__init__(square, number, *args, **kwargs)

    def _add_module(self, module):
        module._parent_column = self
        self._modules.append(module)


class Square(object):
    def __init__(self, duration, proportions, tree_permutation_order, first_multi=(1, 1),
                 reading_direction='horizontal'):

        self._duration = None
        self._proportions = None
        self._modules = {}
        self._tree_permutation_order = None
        self._first_multi = (1, 1)
        self._reading_direction = None
        self._side_size = None

        self.duration = duration
        self.proportions = proportions
        self.reading_direction = reading_direction
        self.tree_permutation_order = tree_permutation_order
        self.first_multi = first_multi
        self._rows = None
        self._columns = None

    @property
    def side_size(self):
        return self._side_size

    @property
    def duration(self):
        if self.modules != {}:
            durations = map(lambda module: module.duration, self.modules.values())
            self._duration = sum(durations)
        return self._duration

    @duration.setter
    def duration(self, value):
        if value is None:
            raise ValueError('duration cannot be None')
        self._duration = value
        self._calculate_module_values()

    @property
    def proportions(self):
        return self._proportions

    @proportions.setter
    def proportions(self, values):
        if values is None:
            raise ValueError('proportions cannot be None')
        if self.side_size is not None:
            if len(values) != self.side_size:
                raise ValueError('wrong proportions length')
        else:
            self._side_size = len(values)
        self._proportions = values
        self._calculate_module_values()

    @property
    def tree_permutation_order(self):
        return self._tree_permutation_order

    @tree_permutation_order.setter
    def tree_permutation_order(self, values):
        if values is None:
            raise ValueError('tree_permutation_order cannot be None')
        if self.side_size is not None:
            if len(values) != self.side_size:
                raise ValueError('wrong tree_permutation_order length')
        else:
            self._side_size = len(values)
        self._tree_permutation_order = values
        self._calculate_module_values()

    @property
    def modules(self):
        return self._modules

    @property
    def first_multi(self):
        return self._first_multi

    @first_multi.setter
    def first_multi(self, value):
        self._first_multi = value
        if self._modules != {}:
            for key in self._modules.keys():
                module = self._modules[key]
                module.multi = self.index_to_r_c(
                    self.r_c_to_index(module.row_number, module.column_number) + self.r_c_to_index(self.first_multi[0],
                                                                                                   self.first_multi[1]))

    @property
    def reading_direction(self):
        return self._reading_direction

    @reading_direction.setter
    def reading_direction(self, val):
        if self._reading_direction:
            raise FractalTreeException('reading_direction can only be set during initialisation')
        permitted = ['horizontal', 'vertical']
        if val not in permitted:
            raise ValueError('reading_direction.value {} must be in {}'.format(val, permitted))
        self._reading_direction = val

    def get_module(self, *args):
        args = tuple(args)
        return self.modules[args]

    def get_all_modules(self):
        return self.modules.values()

    @property
    def rows(self):
        if self._rows is None:
            self._rows = []
            for row_number in range(1, self.side_size + 1):
                row = Row(square=self, number=row_number)
                for column_number in range(1, self.side_size + 1):
                    module = self.get_module(row_number, column_number)
                    row._add_module(module)

                self._rows.append(row)
        return self._rows

    @property
    def columns(self):
        if self._columns is None:
            self._columns = []
            for column_number in range(1, self.side_size + 1):
                column = Column(square=self, number=column_number)
                for row_number in range(1, self.side_size + 1):
                    module = self.get_module(row_number, column_number)
                    column._add_module(module)

                self._columns.append(column)
        return self._columns

    def get_row(self, row_number):
        return self.rows[row_number - 1]

    def get_column(self, column_number):
        return self.columns[column_number - 1]

    def index_to_r_c(self, index):
        row = int(index / self.side_size) % self.side_size + 1
        column = index % self.side_size + 1
        return row, column

    def r_c_to_index(self, row, column):
        index = ((row - 1) * self.side_size) + (column - 1)
        return index

    def _calculate_module_values(self):
        if self.duration is not None and self.proportions is not None and self.tree_permutation_order is not None:
            row_durations = [self.duration * prop / float(sum(self.proportions)) for prop in self.proportions]
            for (row, column) in [(i + 1, j + 1) for i in range(self.side_size) for j in range(self.side_size)]:
                module_durations = [row_durations[row - 1] * prop / float(sum(self.proportions)) for prop in
                                    self.proportions]
                multi = self.index_to_r_c(
                    self.r_c_to_index(row, column) + self.r_c_to_index(self.first_multi[0], self.first_multi[1]))
                module = Module(duration=module_durations[column - 1],
                                tree_permutation_order=self.tree_permutation_order, proportions=self.proportions,
                                multi=multi, reading_direction=self.reading_direction)
                (module.row_number, module.column_number) = (row, column)
                self._modules[(row, column)] = module
                module._parent_square = self

    def change_module_quarter_duration(self, row_number, column_number, new_quarter_duration):
        factor = Fraction(Fraction(new_quarter_duration),
                          Fraction(self.get_module(row_number, column_number).quarter_duration))

        for key in self.modules:
            self.modules[key].duration = self.modules[key].duration * factor

    def change_module_duration(self, row_number, column_number, new_duration, mode='module_duration'):
        if mode == 'module_duration':
            factor = Fraction(Fraction(new_duration), Fraction(self.get_module(row_number, column_number).duration))
        elif mode == 'score_duration':
            factor = Fraction(Fraction(new_duration),
                              Fraction(self.get_module(row_number, column_number).score_duration))
        else:
            raise ValueError()

        for key in self.modules:
            self.modules[key].duration = self.modules[key].duration * factor

    def round_quarter_durations(self):
        for module in self.modules.values():
            module.quarter_duration = round(module.quarter_duration)

    def write_infos(self, text_path, show_row_name=False, show_module_tempo=False,
                    show_quarter_durations=False):
        os.system('touch ' + text_path)
        file = open(text_path, 'w')
        x = PrettyTable(hrules=1)

        column_numbers = [str(number) for number in range(1, self.side_size + 1)]
        if show_row_name:
            x.field_names = ["row", 'row_name', "column:", *column_numbers]
        else:
            x.field_names = ["row", "column:", *column_numbers]

        for row_number in range(1, self.side_size + 1):
            row = self.get_row(row_number)
            if show_row_name:
                row_info = [row_number, row.__name__]
            else:
                row_info = [row_number]

            tempi = [module.tempo for module in row.modules]
            def _round(duration):
                output = round(duration, 1)
                if output != int(output):
                    output = float(output)
                return output
            durations = [_round(module.duration) for module in row.modules]

            if show_module_tempo:
                x.add_row([*row_info, 'tempo', *tempi])

            if show_quarter_durations:
                if None in tempi:
                    raise Exception('set tempi first')
                quarter_durations = [round(float(module.quarter_duration), 2) for module in row.modules]
                x.add_row([*row_info, 'quarter_dur', *quarter_durations])

            x.add_row([*row_info, 'duration', *durations])

        file.write(x.get_string())
        file.close()


from musurgia.agpdf.drawobject import DrawObject
from musurgia.agpdf.font import Font
import copy

from musurgia.agpdf.positioned import Positioned


class Text(DrawObject):
    DEFAULT_FONT_FAMILY = 'Arial'
    DEFAULT_FONT_SIZE = 10
    DEFAULT_FONT_WEIGHT = 'regular'
    DEFAULT_FONT_STYLE = 'regular'

    def __init__(self, text, font_family=None, font_weight=None, font_style=None, font_size=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.font = Font()
        self.font_family = font_family
        self.font_weight = font_weight
        self.font_style = font_style
        self.font_size = font_size
        self._text = None
        self.text = text

    @property
    def font_family(self):
        return self.font.family

    @font_family.setter
    def font_family(self, val):
        if val is None:
            val = self.DEFAULT_FONT_FAMILY
        self.font.family = val

    @property
    def font_size(self):

        return self.font.size

    @font_size.setter
    def font_size(self, val):
        if val is None:
            val = self.DEFAULT_FONT_SIZE
        self.font.size = val

    @property
    def font_weight(self):
        return self.font.weight

    @font_weight.setter
    def font_weight(self, val):
        if val is None:
            val = self.DEFAULT_FONT_WEIGHT
        self.font.weight = val

    @property
    def font_style(self):
        return self.font.style

    @font_style.setter
    def font_style(self, val):
        if val is None:
            val = self.DEFAULT_FONT_STYLE
        self.font.style = val

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, val):
        self._text = str(val)

    def get_relative_x2(self):
        return self.relative_x + len(self.text) * 1000.0 / self.font.size

    def get_relative_y2(self):
        return self.relative_y

    def draw(self, pdf):
        if self.show:
            style = ""
            pdf.set_font(self.font.family, style=style, size=0)
            if self.font.style == 'italic':
                style += 'I'
            if self.font.weight == 'bold':
                style += 'B'
            pdf.set_font(self.font.family, style=style, size=self.font_size)
            pdf.text(x=pdf.x + self.relative_x, y=pdf.y + self.relative_y, txt=self.text)

    def __deepcopy__(self, memodict={}):
        copied = self.__class__(text=self.text)
        for var in vars(self):
            copied_var = copy.deepcopy(vars(self)[var])
            copied.__setattr__(var, copied_var)
        return copied


class TextLabel(Text):
    def __init__(self, text, parent=None, relative_y=-2, *args, **kwargs):
        super().__init__(text=text, relative_y=relative_y, *args, **kwargs)
        self._parent = None
        self.parent = parent
        # self.x_offset = x_offset
        # self.y_offset = y_offset

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, val):
        if val is not None and not isinstance(val, Positioned):
            raise TypeError('parent.value must be of type Positioned not{}'.format(type(val)))
        self._parent = val

    def draw(self, pdf):
        if not self.parent:
            raise Exception('set parent first')
        self.relative_x += self.parent.relative_x
        self.relative_y += self.parent.relative_y
        super().draw(pdf)

    # @DrawObject.relative_x.getter
    # def relative_x(self):
    #     if not self.parent:
    #         raise Exception('set parent first')
    #     return self.parent.relative_x + self.x_offset
    #
    # @DrawObject.relative_y.getter
    # def relative_y(self):
    #     if not self.parent:
    #         raise Exception('set parent first')
    #     return self.parent.relative_y + self.y_offset



import itertools

from quicktions import Fraction

from musurgia.permutation import LimitedPermutation, permute
from musurgia.tree import Tree


class FractalTreeException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class ValueCannotBeChanged(FractalTreeException):
    def __init__(self, *args):
        super().__init__('FractalTree().value can only be changed for the root with no children', *args)


class SetValueFirst(FractalTreeException):
    def __init__(self, *args):
        super().__init__('FractalTree().value must be set before add_layer()', *args)


class FractalTree(Tree):
    def __init__(self, proportions=None, tree_permutation_order=None, value=None, multi=None, fertile=True,
                 reading_direction='horizontal', *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self._value = None
        self._proportions = None
        self._tree_permutation_order = None
        self._multi = None
        self._permutation_order = None
        self._fertile = None
        self._reading_direction = None
        self._fractal_order = None
        self._children_fractal_values = None
        self._name = None

        self.proportions = proportions
        self.tree_permutation_order = tree_permutation_order
        self.value = value
        self.multi = multi
        self.fertile = fertile
        self.reading_direction = reading_direction

    @property
    def proportions(self):
        return self._proportions

    @proportions.setter
    def proportions(self, values):
        if values is None:
            values = [1, 2, 3]
        self._proportions = [Fraction(Fraction(value) / Fraction(sum(values))) for value in values]

    @property
    def tree_permutation_order(self):
        return self._tree_permutation_order

    @tree_permutation_order.setter
    def tree_permutation_order(self, values):
        if values is None:
            values = [3, 1, 2]
        self._tree_permutation_order = values
        self._permutation_order = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if self._value:
            if self.is_root and not self.get_children():
                self._value = val
            else:
                raise ValueCannotBeChanged()
        else:
            self._value = val

        if self._value == 0:
            self.fertile = False

    @property
    def multi(self):
        return self._multi

    @multi.setter
    def multi(self, vals):
        if vals is None:
            vals = (1, 1)

        elif not isinstance(vals, tuple) or len(vals) != 2:
            raise TypeError('multi has to be a tuple with a length of 2')
        self._multi = vals
        self._permutation_order = None

    @property
    def fertile(self):
        return self._fertile

    @fertile.setter
    def fertile(self, val):
        if not isinstance(val, bool):
            raise TypeError('fertile.value must be of type bool not{}'.format(type(val)))
        self._fertile = val

    @property
    def reading_direction(self):
        return self._reading_direction

    @reading_direction.setter
    def reading_direction(self, val):
        if self._reading_direction:
            raise FractalTreeException('reading_direction can only be set during initialisation')
        permitted = ['horizontal', 'vertical']
        if val not in permitted:
            raise ValueError('reading_direction.value {} must be in {}'.format(val, permitted))
        self._reading_direction = val

    @property
    def permutation_order(self):
        def _calculate_permutation_order():
            if self.tree_permutation_order:
                permutation = LimitedPermutation(input_list=list(range(1, len(self.proportions) + 1)),
                                                 main_permutation_order=self.tree_permutation_order, multi=self.multi,
                                                 reading_direction=self.reading_direction)

                return permutation.permutation_order
            else:
                raise FractalTreeException('tree_permutation_order must be set')

        if self._permutation_order is None:
            self._permutation_order = _calculate_permutation_order()
        return self._permutation_order

    @property
    def fractal_order(self):
        return self._fractal_order

    def calculate_position_in_tree(self):
        parent = self.up
        if self.is_root:
            return 0
        else:
            index = parent.get_children().index(self)
            if index == 0:
                return parent.position_in_tree
            else:
                return parent.get_children()[index - 1].position_in_tree + parent.get_children()[index - 1].value

    @property
    def position_in_tree(self):
        return self.calculate_position_in_tree()

    @property
    def number_of_layers(self):
        if len(self.get_leaves()) == 0:
            return 0
        else:
            return self.get_farthest_leaf().get_distance(self)

    @property
    def __name__(self):
        return self.name

    @__name__.setter
    def __name__(self, val):
        self.name = val

    @property
    def name(self):
        if self._name is None:
            if self.is_root:
                self._name = '0'
            elif self.up.is_root:
                self._name = str(self.up.get_children().index(self) + 1)
            else:
                self._name = self.up.name + '.' + str(self.up.get_children().index(self) + 1)
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def position(self):
        if self.is_root():
            # return self.first_position
            return 0
        else:
            siblings = self.up.children
            index = siblings.index(self)
            previous_siblings = siblings[:index]
            position_in_parent = 0
            for child in previous_siblings: position_in_parent += child.value
            return position_in_parent + self.up.position

    @property
    def children_fractal_values(self):
        if not self._children_fractal_values:
            self._children_fractal_values = self._calculate_children_fractal_values()
        return self._children_fractal_values

    def _calculate_children_fractal_orders(self):
        if self.value and self.proportions:
            children_fractal_orders = range(1, len(self.proportions) + 1)
            permutation_order = self.permutation_order
            if permutation_order:
                children_fractal_orders = permute(children_fractal_orders, permutation_order)
            return children_fractal_orders

    @property
    def children_fractal_orders(self):
        return self._calculate_children_fractal_orders()

    def _calculate_children_fractal_values(self):
        value = self.value
        if value == 0:
            value = 0.01
        if value and self.proportions:
            children_fractal_values = [value * prop for prop in self.proportions]
            if self.permutation_order:
                try:
                    children_fractal_values = permute(children_fractal_values, self.permutation_order)
                except ValueError:
                    raise ValueError('proportions and tree_permutation_order should have the same length')
            return children_fractal_values

    def _child_multi(self, parent, index):
        number_of_children = len(self.proportions)
        multi_first = sum(parent.multi) % number_of_children
        if multi_first == 0:
            multi_first = number_of_children
        ch_multi = (multi_first, index + 1)
        return ch_multi

    def add_self(self):
        leaves = self.get_leaves()
        for leaf in leaves:
            new_node = self.copy()
            new_node._up = self
            new_node._fractal_order = self.fractal_order
            leaf.add_child(new_node)

    def add_layer(self, *conditions):
        if self.value is None:
            raise SetValueFirst()

        leaves = list(self.traverse_leaves())
        if not leaves:
            leaves = [self]

        if conditions:
            for leaf in leaves:
                for condition in conditions:
                    if not condition(leaf):
                        leaf.fertile = False
                        break

        for leaf in leaves:
            if leaf.fertile is True:
                for i in range(len(leaf.proportions)):
                    new_node = leaf.copy()
                    new_node.value = leaf.children_fractal_values[i]
                    new_node.multi = self._child_multi(leaf, i)
                    leaf.add_child(new_node)
                    new_node._fractal_order = leaf.children_fractal_orders[i]

            else:
                pass

    def get_layer(self, layer, key=None):
        if layer <= self.get_root().number_of_layers:
            branch_distances = []
            for child in self.get_children():
                branch_distances.append(child.get_farthest_leaf().get_distance())
            if layer == 0:
                if key:
                    return [getattr(self, key)]
                else:
                    return [self]

            if layer >= 1:
                output = []
                if not branch_distances:
                    output.extend(self.get_layer(layer=layer - 1, key=key))
                for i in range(len(self.get_children())):
                    child = self.get_children()[i]
                    if branch_distances[i] == 1:
                        if key:
                            output.append(getattr(child, key))
                        else:
                            output.append(child)
                    else:
                        if layer == 1:
                            output.extend(child.get_layer(layer - 1, key))
                        else:
                            output.append(child.get_layer(layer - 1, key))

                return output
        else:
            err = 'max layer number=' + str(self.number_of_layers)
            raise ValueError(err)

    @property
    def layers(self):
        return [self.get_layer(i) for i in range(self.number_of_layers + 1)]

    @property
    def next(self):
        if not self.is_leaf():
            raise Exception('FractalTree().next property can only be used for leaves')
        else:
            try:

                parent = self.up
                while parent is not None:
                    parent_leaves = list(parent.traverse_leaves())
                    index = parent_leaves.index(self)
                    try:
                        return parent_leaves[index + 1]
                    except IndexError:
                        parent = parent.up
                raise IndexError()
            except IndexError:
                return None

    def get_previous_leaf(self):
        if not self.is_leaf():
            raise Exception('FractalTree().get_previous_leaf  can only be applied to leaves')
        try:
            parent = self.up
            while parent is not None:
                parent_leaves = list(parent.traverse_leaves())
                index = parent_leaves.index(self)
                try:
                    return parent_leaves[index - 1]
                except IndexError:
                    parent = parent.up
            raise IndexError()
        except IndexError:
            return None

    def get_previous_sibling(self):
        try:
            siblings = self.up.get_children()
            index = siblings.index(self)
            if index == 0:
                return None
            return siblings[index - 1]
        except AttributeError:
            return None

    def get_next_sibling(self):
        try:
            siblings = self.up.get_children()
            index = siblings.index(self)
            return siblings[index + 1]
        except (IndexError, AttributeError):
            return None

    def split(self, *proportions):

        for prop in proportions:
            self.add_self()
            self.get_children()[-1].value = self.value * prop / sum(proportions)

    def round_leaves(self, min_diff=1):
        for leaf in self.traverse_leaves():
            factor = 1 / min_diff
            new_value = round(leaf.value * factor) / factor
            leaf._value = Fraction(new_value)

        difference = self.value - sum([leaf.value for leaf in self.traverse_leaves()])
        # if difference != 0:
        #     warnings.warn('tree node {} round leaves difference= {}'.format(self.__name__, difference))
        return difference

    def _reduce(self):
        sisters = self.get_siblings()

        if not sisters:
            raise Exception('no reduction possible without sisters')
        else:

            non_zero_sisters = [sister for sister in sisters if sister.value != 0]

            if non_zero_sisters:
                addition = Fraction(self.value, len(non_zero_sisters))
                for sister in non_zero_sisters:
                    if sister.value != 0:
                        sister._value += addition
                self._value = 0

    def reduce_children(self, condition):

        for child in self.get_children():
            if condition(child):
                child._reduce()

        to_be_detached = [child for child in self.get_children() if condition(child)]
        for child in to_be_detached:
            child.detach()

        self._children_fractal_values = [child.value for child in self.get_children()]

    def _check_merge_nodes(self, nodes):
        return True

    def merge_children(self, *lengths):
        children = self.get_children()
        if not children:
            raise Exception('there are no children to be merged')
        if sum(lengths) != len(children):
            raise ValueError(
                'sum of lengths {} must be the same as length of children {}'.format(sum(lengths), len(children)))

        def _merge(nodes):
            self._check_merge_nodes(nodes)

            node_values = [node.value for node in nodes]

            nodes[0]._value = sum(node_values)
            for node in nodes[1:]:
                self.remove_child(node)

        iter_children = iter(children)
        chunks = [list(itertools.islice(iter_children, l)) for l in lengths]

        for chunk in chunks:
            _merge(chunk)

    def copy(self):
        return self.__class__(value=self.value, proportions=self.proportions,
                              tree_permutation_order=self.tree_permutation_order, multi=self.multi,
                              reading_direction=self.reading_direction, fertile=self.fertile)

    def __deepcopy__(self, memodict={}):
        copied = self.copy()
        copied._fractal_order = self.fractal_order
        copied.__name__ = self.__name__
        return copied



import os
from unittest import TestCase

from musurgia.testcomparefiles import TestCompareFiles

tfc = TestCompareFiles()

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def test_1_1(self):
        with self.assertRaises(ValueError):
            wrong_path = path + '_test_1'
            tfc.assertTemplate(wrong_path)

    def test_1_2(self):
        file_path = path + '_test_1.txt'
        tfc.assertTemplate(file_path)

    def test_2_1(self):
        file_path = path + '_test_2.txt'
        with self.assertRaises(AssertionError):
            tfc.assertTemplate(file_path)

    def test_2_2(self):
        file_path = path + '_test_2.txt'
        template_path= path + '_test_1_template.txt'
        tfc.assertTemplate(file_path=file_path, template_path=template_path)



import os
from unittest import TestCase

from musurgia.agpdf.pdf import Pdf, PageText
from fpdf import FPDF

from musurgia.testcomparefiles import TestCompareFiles

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_0_1(self):
        pdf = self.pdf
        pdf_path = path + '_test_0_1.pdf'
        pdf.y = 100
        t = PageText('bold_test', font_weight='bold')
        t.draw(pdf)
        pdf.y = 200
        t = PageText('normal_test')
        t.draw(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_0_2(self):
        pdf = self.pdf
        pdf_path = path + '_test_0_2.pdf'
        pdf.y = 100
        t = PageText('large_test', font_size=15)
        t.draw(pdf)
        pdf.y = 200
        t = PageText('normal_test')
        t.draw(pdf)
        pdf.write(pdf_path)
        # TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_1(self):
        pdf = self.pdf
        pdf_path = path + '_test_1.pdf'
        pdf.y = 100
        t = PageText('hallo world!', font_weight='bold')
        t.draw(pdf)
        pdf.add_page()
        t.draw(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_2(self):
        pdf = self.pdf
        pdf.add_page()

        pdf_path = path + '_test_2.pdf'
        pdf.page = 1
        pdf.y = 100
        t = PageText('hallo world!', font_weight='bold')

        t.draw(pdf)
        # t = PageText('hallo new world!', font_weight='bold')
        pdf.page = 2
        # print(pdf.font_style)
        pdf.y = 100
        t.draw(pdf)
        # print(pdf.font_style)

        pdf.page = 1
        pdf.y = 150
        t.draw(pdf)
        pdf.page = 2
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_3(self):
        pdf = self.pdf
        pdf.add_page()

        pdf_path = path + '_test_3.pdf'
        pdf.page = 1
        pdf.y = 100
        t = PageText('hallo world!')

        t.draw(pdf)
        t.font_weight = 'bold'
        # t = PageText('hallo new world!', font_weight='bold')
        pdf.page = 2

        pdf.y = 100
        t.draw(pdf)
        # print(pdf.font_style)

        pdf.page = 1
        pdf.y = 150
        t.draw(pdf)
        pdf.page = 2
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_4(self):
        pdf = FPDF()
        pdf.font_family = 'Arial'
        pdf.add_page()
        pdf.add_page()

        pdf_path = path + '_test_4.pdf'
        pdf.page = 1
        pdf.text(10, 100, 'hello world!')

        pdf.page = 2
        pdf.set_font(family=pdf.font_family, style='B')
        pdf.text(10, 100, 'hello new world!')

        pdf.page = 1
        # print(pdf.font_style)
        # pdf.set_font(family=pdf.font_family, style='BI')
        pdf.set_font(family=pdf.font_family, style='')
        pdf.set_font(family=pdf.font_family, style='B')
        pdf.text(10, 150, 'hello old world!')
        pdf.page = 2
        pdf.output(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)



from unittest import TestCase
import os

from diff_pdf_visually import pdfdiff


class TestCompareFiles(TestCase):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _compare_pdfs(self, file_path, template_path, verbosity):
        self.assertTrue(pdfdiff(file_path, template_path, verbosity=verbosity))

    def _compare_contents(self, file_path, template_path):
        with open(file_path, 'r') as myfile:
            result = myfile.read()

        with open(template_path, 'r') as myfile:
            template = myfile.read()

        self.assertEqual(template, result)

    def assertTemplate(self, file_path, template_path=None, verbosity=0):
        file_name, extension = os.path.splitext(file_path)
        if not template_path:
            if not extension:
                raise ValueError('file_path has no file extension')
            template_path = file_name + '_template' + extension

        if extension == '.pdf':
            self._compare_pdfs(file_path, template_path, verbosity=verbosity)
        else:
            self._compare_contents(file_path, template_path)



import os
from unittest import TestCase

from musurgia.agpdf.line import Line
from musurgia.agpdf.linegroup import LineGroup
from musurgia.agpdf.pdf import Pdf

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_1(self):
        pdf = self.pdf
        pdf_path = path + '_test_1.pdf'
        lg = LineGroup(inner_distance=7)
        lg.add_line(Line(length=10))
        lg.add_line(Line(length=10))
        lg.add_line(Line(length=10))

        lg.draw(pdf)
        pdf.write(pdf_path)

    def test_2(self):
        def make_line_group():
            lg = LineGroup(inner_distance=7, bottom_distance=30)
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            return lg

        pdf = self.pdf
        pdf_path = path + '_test_2.pdf'

        lg = make_line_group()
        lg.draw(pdf)
        pdf.write(pdf_path)

    def test_3(self):
        def make_line_group():
            lg = LineGroup(inner_distance=7, bottom_distance=30)
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            return lg

        pdf = self.pdf
        pdf_path = path + '_test_3.pdf'
        lg = make_line_group()
        lg.draw(pdf)

        lg = make_line_group()
        lg.draw(pdf)
        pdf.write(pdf_path)

    def test_4(self):
        def make_line_group():
            lg = LineGroup(inner_distance=7, bottom_distance=30)
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            return lg

        pdf = self.pdf
        pdf_path = path + '_test_4.pdf'
        lg = make_line_group()
        lg.draw(pdf)

        lg = make_line_group()
        lg.relative_x = 50
        lg.relative_y = 50
        lg.draw(pdf)
        pdf.write(pdf_path)

    def test_5a(self):
        pdf = self.pdf
        pdf_path = path + '_test_5a.pdf'
        line = Line(length=10)
        line.draw_with_break(pdf)

        line = Line(length=10)
        line.relative_x = 300
        line.draw_with_break(pdf)
        pdf.write(pdf_path)

    def test_5(self):
        def make_line_group():
            lg = LineGroup(inner_distance=7, bottom_distance=30)
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            return lg

        pdf = self.pdf
        pdf_path = path + '_test_5.pdf'
        lg = make_line_group()
        lg.draw_with_break(pdf)

        lg = make_line_group()
        lg.relative_x = 300
        lg.draw_with_break(pdf)
        pdf.write(pdf_path)

    def test_6(self):
        def make_line_group():
            lg = LineGroup(inner_distance=7, bottom_distance=30)
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            lg.add_line(Line(length=10))
            return lg

        pdf = self.pdf
        pdf_path = path + '_test_6.pdf'
        lg = make_line_group()
        lg.draw_with_break(pdf)

        lg = make_line_group()
        lg.relative_x = 300
        lg.relative_y = 240
        lg.draw_with_break(pdf)
        pdf.write(pdf_path)



import os
from unittest import TestCase

from musicscore.musictree.midi import C

from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.testcomparefiles import TestCompareFiles

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(proportions=[1, 2, 3, 4], tree_permutation_order=[3, 1, 4, 2], quarter_duration=20,
                               tempo=70)
        self.fm.midi_generator.midi_range = [36, 60]

    def test_1(self):
        self.fm.add_layer()
        score = self.fm.get_score(layer_number=1)

        for leaf in self.fm.traverse_leaves():
            leaf.set_chord(leaf.get_choral(range_factor=1.5, direction=1, last=True))

        sf = self.fm.get_simple_format()

        v = sf.to_stream_voice(1)
        v.add_to_score(score, 1, 2)

        xml_path = path + '_test_1.xml'
        score.write(path=xml_path)

        TestCompareFiles().assertTemplate(file_path=xml_path)

    def test_2(self):
        self.fm.midi_generator.midi_range = [C(4).value, C(6).value]
        self.fm.midi_generator.directions = [1, -1, 1]
        self.fm.add_layer()
        score = self.fm.get_score(layer_number=1)

        for leaf in self.fm.traverse_leaves():
            leaf.set_chord(leaf.get_choral(range_factor=1, direction=1, last=False))
        sf = self.fm.get_simple_format()

        v = sf.to_stream_voice(1)
        v.add_to_score(score, 1, 2)

        for leaf in self.fm.traverse_leaves():
            leaf.add_layer()

        sf = self.fm.get_simple_format()

        v = sf.to_stream_voice(1)
        v.add_to_score(score, 1, 3)

        xml_path = path + '_test_2.xml'
        score.write(path=xml_path)

        TestCompareFiles().assertTemplate(file_path=xml_path)



from musurgia.agpdf.drawobject import DrawObject
from musurgia.agpdf.linegroup import LineGroup
from musurgia.timeline.ruler import Ruler
from musurgia.timeline.voice import Voice
from musurgia.tree import Tree


class TimeLineError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TimeLine(DrawObject, Tree):
    def __init__(self, length=None, unit=10, inner_distance=7, bottom_distance=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._length = None
        self.length = length
        self._unit = None
        self.unit = unit
        self._inner_distance = None
        self.inner_distance = inner_distance
        self._bottom_distance = None
        self.bottom_distance = bottom_distance
        self._line_groups = None

    @property
    def ruler(self):
        try:
            return self.get_children()[0]
        except IndexError:
            return None

    @property
    def voices(self):
        try:
            return self.get_children()[1:]
        except IndexError:
            return []

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, val):
        if val is not None:
            if self._length:
                raise TimeLineError('length can only be set once')
            self.add_child(Ruler(length=val, unit=10))

        self._length = val

    @property
    def unit(self):
        return self._unit

    def update_units(self):
        for child in self.get_children():
            child.update_unit(self.unit)

    @unit.setter
    def unit(self, val):
        self._unit = val
        self.update_units()

    @property
    def inner_distance(self):
        return self._inner_distance

    @inner_distance.setter
    def inner_distance(self, val):
        self._inner_distance = val

    @property
    def bottom_distance(self):
        return self._bottom_distance

    @bottom_distance.setter
    def bottom_distance(self, val):
        self._bottom_distance = val

    def add_voice(self, name):
        if self._line_groups:
            raise Exception()
        voice = Voice(length=self.length, unit=self.unit, name=name)
        return self.add_child(voice)

    @property
    def line_groups(self):
        if self._line_groups is None:
            output = []
            for index, line in enumerate(self.ruler.lines):
                gl = LineGroup(inner_distance=self.inner_distance, bottom_distance=self.bottom_distance)
                gl.add_line(line)
                for voice in self.voices:
                    gl.add_line(voice.lines[index])
                output.append(gl)
            self._line_groups = output
        return self._line_groups

    def get_relative_x2(self):
        raise Exception('timeline has no x2 value')

    def get_relative_y2(self):
        return self.line_groups[0].get_relative_y2()

    def get_height(self):
        return self.get_relative_y2() - self.line_groups[0].relative_y

    def draw(self, pdf):
        for line_group in self.line_groups:
            if line_group == self.line_groups[0]:
                for child, line in zip(self.get_children()[1:], line_group.lines[1:]):
                    line.name = child.name
                    if line.name:
                        line.name.relative_y = line.relative_y

            line_group.draw_with_break(pdf)
            new_x = pdf.x

            if line_group._line_break:
                for child, line in zip(self.get_children(), line_group.lines):
                    pdf.x = new_x - line_group.length
                    line.name = child.name

                    if line.name:
                        line.name.draw(pdf)
                pdf.x = new_x



import os
from unittest import TestCase

from musurgia.agpdf.pdf import Pdf
from musurgia.agpdf.textlabel import Text
from musurgia.testcomparefiles import TestCompareFiles

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_1(self):
        pdf_path = path + '_test_1.pdf'
        tl = Text('test_1')
        tl.relative_x = 30
        copied = tl.__deepcopy__()
        copied.draw(self.pdf)
        self.pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)



import os
from unittest import TestCase

from musurgia.agpdf.pdf import Pdf, PageText
from musicscore.musictree.treeinstruments import Violin, Viola, Cello, Percussion, Accordion

from musurgia.fractaltree.fractalmusicsquare import Square
from musurgia.testcomparefiles import TestCompareFiles
from musurgia.timeline.scoretimeline import ScoreTimeLine, ModuleTimeLine

path = str(os.path.abspath(__file__).split('.')[0])

square = Square(duration=100, tree_permutation_order=[3, 1, 2], proportions=[1, 2, 3])

square.change_module_duration(1, 1, 4)
square.get_row(1).set_tempo(50)
square.get_row(2).set_tempo(80)
square.get_row(3).set_tempo(100)

for module in square.modules.values():
    module.quarter_duration = round(module.quarter_duration)

# square.get_row(1).set_score_tempo(50)
# square.get_row(1).set_module_tempo(72)
# square.get_row(2).set_score_tempo(80)
# square.get_row(2).set_module_tempo(72)
# square.get_row(3).set_score_tempo(100)
# square.get_row(3).set_module_tempo(72)


VIOLIN = Violin()
VIOLA = Viola()
CELLO = Cello()
PERCUSSION = Percussion()
ACCORDION = Accordion()


class Test(TestCase):

    def test_1(self):
        pdf = Pdf(orientation='landscape', t_margin=25, l_margin=20)
        title = PageText(text='t e s t', v_position='center', h_position='top', relative_y=-14, font_size=18,
                         font_weight='bold')
        composer = PageText(text='A. G. 2019', v_position='right', h_position='top', relative_y=-14)

        pdf_path = path + '_test_1.pdf'
        stl = ScoreTimeLine(instruments=[VIOLIN, VIOLA, CELLO, PERCUSSION, ACCORDION], units_per_line=60)
        stl.add_module_time_line(
            ModuleTimeLine(start_time=0, module=square.get_module(1, 2), instruments=[VIOLIN, VIOLA, CELLO]))
        stl.add_module_time_line(
            ModuleTimeLine(start_time=16, module=square.get_module(2, 3), instruments=[PERCUSSION]))
        stl.add_module_time_line(
            ModuleTimeLine(start_time=20, module=square.get_module(3, 1), instruments=[ACCORDION]))
        stl.add_module_time_line(
            ModuleTimeLine(start_time=60, module=square.get_module(3, 3),
                           instruments=[VIOLIN, VIOLA, CELLO, PERCUSSION, ACCORDION]))

        stl.add_module_time_line(
            ModuleTimeLine(start_time=120, module=square.get_module(1, 3),
                           instruments=[VIOLIN, VIOLA, CELLO]))

        stl.add_module_time_line(
            ModuleTimeLine(start_time=135, module=square.get_module(3, 3),
                           instruments=[ACCORDION, PERCUSSION]))

        stl.add_module_time_line(
            ModuleTimeLine(start_time=160, module=square.get_module(3, 3),
                           instruments=[ACCORDION, PERCUSSION]))

        stl.draw(pdf)
        for page in pdf.pages:
            pdf.page = page
            title.draw(pdf)
            composer.draw(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)
